#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase

from ..interception import Interceptor
from ..call import Types, Curried, Retries, Memoize


class CallTests(UTCase):

    def _assertCall(self, f, *args, **kwargs):

        self.assertRaises(Exception, f, *args, **kwargs)

    def testTypes(self):

        @Types(rtype=int)
        def a(p=None):
            return p

        a(1)
        a()
        self._assertCall(a, '')

        @Types(rtype=Types.NotNone(int))
        def b(p=None):
            return p

        b(2)
        self._assertCall(b, '')
        self._assertCall(b)

        @Types(rtype=[int])
        def c(p=None):
            return [2, 3] \
                if p == 1 else \
                None if p == 2 else \
                [] if p == 3 else \
                [2, None] if p == 4 else [2, '']

        c(1)
        c(2)
        c(3)
        c(4)
        self._assertCall(c)

        @Types([Types.NotNone(int)])
        def d(p=None):
            return [2, 3] \
                if p == 1 else \
                None if p == 2 else \
                [] if p == 3 else \
                [2, None] if p == 4 else [2, '']

        d(1)
        d(2)
        d(3)
        self._assertCall(d, 4)
        self._assertCall(d)

        @Types(Types.NotEmpty([int]))
        def e(p=None):
            return [2, 3] \
                if p == 1 else \
                None if p == 2 else \
                [] if p == 3 else \
                [2, None] if p == 4 else [2, '']

        e(1)
        e(2)
        self._assertCall(e, 3)
        e(4)
        self._assertCall(e)

        @Types(Types.NotEmpty([Types.NotNone(int)]))
        def f(p=None):
            return [2, 3] \
                if p == 1 else \
                None if p == 2 else \
                [] if p == 3 else \
                [2, None] if p == 4 else [2, '']

        f(1)
        f(2)
        self._assertCall(f, 3)
        self._assertCall(f, 4)
        self._assertCall(f)

        @Types(Types.NotNone(Types.NotEmpty([int])))
        def g(p=None):
            return [2, 3] \
                if p == 1 else \
                None if p == 2 else \
                [] if p == 3 else \
                [2, None] if p == 4 else [2, '']

        g(1)
        self._assertCall(g, 2)
        self._assertCall(g, 3)
        g(4)
        self._assertCall(g)

        @Types(ptypes={'p': int})
        def f(p=None):
            pass

        a(1)
        a()
        self._assertCall(a, '')

        @Types(ptypes={'p': Types.NotNone(int)})
        def g(p=None):
            pass

        g(1)
        self._assertCall(b)
        self._assertCall(b, '')

        @Types(ptypes={'p': [int]})
        def h(p=None):
            pass

        h([1])
        h([2, 3])
        h([None])
        h([])
        h()
        self._assertCall(h, [2, ''])

        @Types(ptypes={'p': [Types.NotNone(int)]})
        def i(p=None):
            pass

        i([1])
        i([2, 3])
        self._assertCall(i, [None])
        i([])
        i()
        self._assertCall(i, [2, ''])

        @Types(ptypes={'p': Types.NotEmpty([int])})
        def d(p=None):
            pass

        d()
        d([1])
        d([2, 3])
        self._assertCall(d, 3)
        self._assertCall(d, [])
        self._assertCall(d, [2, ''])

    def testInterceptor(self):

        def interception(target, args, kwargs):
            raise Exception()

        @Interceptor(interception)
        def a(a, b=2):
            pass

        self._assertCall(a, a=None)
        self._assertCall(a)

    def testCurried(self):

        @Curried()
        def a(b, c=None):
            return 2

        result = a()

        self.assertTrue(isinstance(result, Curried.CurriedResult))

        self.assertEqual(a(None), 2)

        result.curried.args = []

        self.assertEqual(a(b=None), 2)

        self.assertEqual(a(b=None, c=None), 2)

        self.assertTrue(isinstance(a(None), Curried.CurriedResult))

        pass

    def testRetries(self):

        global count
        count = 10

        @Retries(10, delay=0, backoff=0)
        def a():
            global count
            count -= 1
            if count > 0:
                raise Exception()
            else:
                return ""

        result = a()

        self.assertTrue(count == 0)
        self.assertTrue(result == "")


class MemoizeTest(UTCase):
    """Test the memoize annotation."""

    def setUp(self):

        self.memoize = Memoize()

        self.n = 0

        def func(*args, **kwargs):

            self.n += 1

            return self.n

        self.func = self.memoize(func)

    def test_empty(self):

        result = self.func()
        self.assertEqual(result, 1)

        params = self.memoize.getparams(1)
        self.assertEqual(params, ((), {}))

        result = self.func()
        self.assertEqual(result, 1)

    def test_maxsize(self):

        self.memoize.max_size = 2

        result = self.func()
        self.assertEqual(result, 1)

        params = self.memoize.getparams(1)
        self.assertEqual(params, ((), {}))

        result = self.func()
        self.assertEqual(result, 1)

        params = self.memoize.getparams(1)
        self.assertEqual(params, ((), {}))

        result = self.func(1, 2, a=3)
        self.assertEqual(result, 2)

        params = self.memoize.getparams(2)
        self.assertEqual(params, ((1, 2), {'a': 3}))

        result = self.func(3, 4, b=5)
        self.assertEqual(result, 3)

        self.assertRaises(ValueError, self.memoize.getparams, 3)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Definition of annotation dedicated to intercept annotated element calls."""

from .core import Annotation

from b3j0f.aop import weave, unweave

__all__ = [
    'Interceptor',
    'PrivateInterceptor', 'CallInterceptor', 'PrivateCallInterceptor'
]


class Interceptor(Annotation):
    """Annotation able to intercept annotated elements.

    This interception can be disabled at any time and specialized with a
        pointcut.
    """

    #: interception attribute name
    INTERCEPTION = 'interception'

    #: pointcut attribute name
    POINTCUT = 'pointcut'

    #: attribute name for enable the interception
    ENABLE = 'enable'

    #: interceptor attribute name
    INTERCEPTOR = '__interceptor__'

    #: private attribute name for pointcut
    _POINTCUT = '_pointcut'

    __slots__ = (
        INTERCEPTION, ENABLE,  # public attributes
        _POINTCUT  # private attribute
    ) + Annotation.__slots__

    class InterceptorError(Exception):
        """Handle Interceptor errors."""

    def __init__(
            self, interception=None, pointcut=None, enable=True,
            *args, **kwargs
    ):
        """Default constructor with interception function and enable property.

        :param callable interception: called if self is enabled and if any
            annotated element is called. Parameters of call are self annotation
            and an AdvicesExecutor.
        :param pointcut: pointcut to use in order to weave interception.
        :param bool enable:
        """

        super(Interceptor, self).__init__(*args, **kwargs)

        self.interception = interception
        self._pointcut = pointcut
        self.enable = enable

    @property
    def pointcut(self):
        """Get pointcut."""

        return self._pointcut

    @pointcut.setter
    def pointcut(self, value):
        """Change of pointcut.
        """

        pointcut = getattr(self, Interceptor.POINTCUT)

        # for all targets
        for target in self.targets:
            # unweave old advices
            unweave(target, pointcut=pointcut, advices=self.intercepts)
            # weave new advices with new pointcut
            weave(target, pointcut=value, advices=self.intercepts)

        # and save new pointcut
        setattr(self, Interceptor._POINTCUT, value)

    def _bind_target(self, target, ctx=None, *args, **kwargs):
        """Weave self.intercepts among target advices with pointcut."""

        result = super(Interceptor, self)._bind_target(
            target=target, ctx=ctx, *args, **kwargs
        )

        pointcut = getattr(self, Interceptor.POINTCUT)

        weave(result, pointcut=pointcut, advices=self.intercepts, ctx=ctx)

        return result

    def remove_from(self, target, ctx=None, *args, **kwargs):

        super(Interceptor, self).remove_from(target, *args, **kwargs)

        unweave(
            target, pointcut=self.pointcut, advices=self.intercepts, ctx=ctx
        )

    def intercepts(self, joinpoint):
        """Self target interception if self is enabled

        :param joinpoint: advices executor
        """

        result = None

        if self.enable:

            interception = getattr(self, Interceptor.INTERCEPTION)

            joinpoint.exec_ctx[Interceptor.INTERCEPTION] = self

            result = interception(joinpoint)

        else:
            result = joinpoint.proceed()

        return result

    @classmethod
    def set_enable(cls, target, enable=True):
        """(Dis|En)able annotated interceptors."""

        interceptors = cls.get_annotations(target)

        for interceptor in interceptors:
            setattr(interceptor, Interceptor.ENABLE, enable)


class PrivateInterceptor(Interceptor):
    """Interceptor with a private interception resource.
    """

    __slots__ = Interceptor.__slots__

    def __init__(self, *args, **kwargs):
        """
        Do nothing except set self._interception such as its interception
        """
        super(PrivateInterceptor, self).__init__(
            interception=self._interception, *args, **kwargs
        )

    def _interception(self, joinpoint):
        """
        Default interception which raises a NotImplementedError.

        :param Annotation annotation: intercepted annotation
        """

        raise NotImplementedError()


class CallInterceptor(Interceptor):
    """Interceptor dedicated to intercept call of annotated element.

    Instead of Interceptor, the pointcut equals '__call__', therefore, the
    target may be a class instead of an instance beceause python does not
    handle system methods (such as __call__, __repr__, etc.) if they are
    overriden on instance.

    It could be used to intercepts annotation binding annotated by self.
    """

    __CALL__ = '__call__'

    __slots__ = Interceptor.__slots__

    def __init__(self, *args, **kwargs):

        super(CallInterceptor, self).__init__(
            pointcut=CallInterceptor.__CALL__, *args, **kwargs
        )


class PrivateCallInterceptor(CallInterceptor):
    """Interceptor dedicated to apply a private interception on target calls.

    Instead of Interceptor, the pointcut equals '__call__', therefore, the
    target may be a class instead of an instance beceause python does not
    handle system methods (such as __call__, __repr__, etc.) if they are
    overriden on instance.
    """

    __slots__ = CallInterceptor.__slots__ + PrivateInterceptor.__slots__

    def __init__(self, *args, **kwargs):

        super(PrivateCallInterceptor, self).__init__(
            interception=self._interception, *args, **kwargs
        )

    def _interception(self, joinpoint):
        """
        Default interception which raise a NotImplementedError.
        """

        raise NotImplementedError()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Decorators dedicated to asynchronous programming."""

from __future__ import absolute_import

try:
    from threading import Thread, RLock
except ImportError:
    from dummythreading import Thread, RLock

from time import sleep

from signal import signal, SIGALRM, alarm

from six import callable
from six.moves.queue import Queue

from .core import Annotation
from .interception import PrivateInterceptor
from .oop import Mixin

__all__ = [
    'Synchronized', 'SynchronizedClass',
    'Asynchronous', 'TimeOut', 'Wait', 'Observable'
]


class Synchronized(PrivateInterceptor):
    """Transform a target into a thread safe target."""

    #: lock attribute name
    _LOCK = '_lock'

    __slots__ = (_LOCK,) + PrivateInterceptor.__slots__

    def __init__(self, lock=None, *args, **kwargs):

        super(Synchronized, self).__init__(*args, **kwargs)

        self._lock = RLock() if lock is None else lock

    def _interception(self, joinpoint):

        self._lock.acquire()

        result = joinpoint.proceed()

        self._lock.release()

        return result


class SynchronizedClass(Synchronized):
    """Transform a class into a thread safe class."""

    def on_bind_target(self, target, ctx=None):

        for attribute in target.__dict__:
            if callable(attribute):
                Synchronized(attribute, self._lock)


class Asynchronous(Annotation):
    """Transform a target into an asynchronous callable target."""

    def __init__(self, *args, **kwargs):

        super(Asynchronous, self).__init__(*args, **kwargs)

        self.queue = None

    def _threaded(self, *args, **kwargs):
        """Call the target and put the result in the Queue."""

        for target in self.targets:
            result = target(*args, **kwargs)
            self.queue.put(result)

    def on_bind_target(self, target, ctx=None):

        # add start function to wrapper
        super(Asynchronous, self).on_bind_target(target, ctx=ctx)

        setattr(target, 'start', self.start)

    def start(self, *args, **kwargs):
        """Start execution of the function."""

        self.queue = Queue()
        thread = Thread(target=self._threaded, args=args, kwargs=kwargs)
        thread.start()

        return Asynchronous.Result(self.queue, thread)

    class NotYetDoneException(Exception):
        """Handle when a result is not yet available."""

    class Result(object):
        """In charge of receive asynchronous function result."""

        __slots__ = ('queue', 'thread', 'result')

        def __init__(self, queue, thread):

            super(Asynchronous.Result, self).__init__()

            self.result = None
            self.queue = queue
            self.thread = thread

        def is_done(self):
            """True if result is available."""

            return not self.thread.is_alive()

        def get_result(self, wait=-1):
            """Get result value.

            Wait for it if necessary.

            :param int wait: maximum wait time.
            :return: result value.
            """

            if not self.is_done():

                if wait >= 0:
                    self.thread.join(wait)

                else:
                    raise Asynchronous.NotYetDoneException(
                        'the call has not yet completed its task'
                    )

            if self.result is None:
                self.result = self.queue.get()

            return self.result


class TimeOut(PrivateInterceptor):
    """Raise an Exception if the target call has not finished in time."""

    class TimeOutError(Exception):
        """Exception thrown if time elapsed before the end of the target call.
        """

        #: Default time out error message.
        DEFAULT_MESSAGE = \
            'Call of {0} with parameters {1} and {2} is timed out in frame {3}'

        def __init__(self, timeout_interceptor, frame):

            super(TimeOut.TimeOutError, self).__init__(
                timeout_interceptor.message.format(
                    timeout_interceptor.target,
                    timeout_interceptor.args,
                    timeout_interceptor.kwargs,
                    frame
                )
            )

    SECONDS = 'seconds'
    ERROR_MESSAGE = 'error_message'

    __slots__ = (SECONDS, ERROR_MESSAGE) + PrivateInterceptor.__slots__

    def __init__(
            self,
            seconds, error_message=TimeOutError.DEFAULT_MESSAGE,
            *args, **kwargs
    ):

        super(TimeOut, self).__init__(*args, **kwargs)

        self.seconds = seconds
        self.error_message = error_message

    def _handle_timeout(self, frame=None, **_):
        """Sig ALARM timeout function."""

        raise TimeOut.TimeOutError(self, frame)

    def _interception(self, joinpoint):

        signal(SIGALRM, self._handle_timeout)
        alarm(self.seconds)

        try:
            result = joinpoint.proceed()

        finally:
            alarm(0)

        return result


class Wait(PrivateInterceptor):
    """Define a time to wait before and after a target call."""

    DEFAULT_BEFORE = 1  #: default seconds to wait before the target call.
    DEFAULT_AFTER = 1  #: default seconds to wait after the target call.

    BEFORE = 'before'  #: before attribute name.

    AFTER = 'after'  #: after attribute name.

    __slots__ = (BEFORE, AFTER) + PrivateInterceptor.__slots__

    def __init__(
            self, before=DEFAULT_BEFORE, after=DEFAULT_AFTER, *args, **kwargs
    ):

        super(Wait, self).__init__(*args, **kwargs)

        self.before = before
        self.after = after

    def _interception(self, joinpoint):

        sleep(self.before)

        result = joinpoint.proceed()

        sleep(self.after)

        return result


class Observable(PrivateInterceptor):
    """Imlementation of the observer design pattern.

    It transforms a target into an observable object in adding method
    register_observer, unregister_observer and notify_observers.
    Observers listen to pre/post target interception.
    """

    def __init__(self, *args, **kwargs):

        super(Observable, self).__init__(*args, **kwargs)

        self.observers = set()

    def register_observer(self, observer):
        """Register an observer."""

        self.observers.add(observer)

    def unregister_observer(self, observer):
        """Unregister an observer."""

        self.observers.remove(observer)

    def notify_observers(self, joinpoint, post=False):
        """Notify observers with parameter calls and information about
        pre/post call.
        """

        _observers = tuple(self.observers)

        for observer in _observers:
            observer.notify(joinpoint=joinpoint, post=post)

    def on_bind_target(self, target, ctx=None):

        Mixin.set_mixin(target, self.register_observer)
        Mixin.set_mixin(target, self.unregister_observer)
        Mixin.set_mixin(target, self.notify_observers)

    def _interception(self, joinpoint):

        self.notify_observers(joinpoint=joinpoint)

        result = joinpoint.proceed()

        self.notify_observers(joinpoint=joinpoint, post=True)

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from __future__ import absolute_import

from unittest import main

from b3j0f.utils.ut import UTCase

from six.moves import range

from ..core import Annotation
from ..check import Condition, MaxCount, Target, AnnotationChecker

from random import randint


class ConditionTest(UTCase):
    """Test the Condition annotation.
    """
    def setUp(self):

        self.pre_count = 0
        self.post_count = 0
        self.condition = Condition(
            pre_cond=self.pre_cond, post_cond=self.post_cond
        )
        self.condition(self._test)

    def pre_cond(self, joinpoint):

        self.pre_count += 1

    def post_cond(self, joinpoint):

        self.post_count += 1

    def tearDown(self):

        self.condition.__del__()
        del self.condition

    def _test(self, **kwargs):
        """Test method which is intercepted by this condition.
        """

        return self

    def test_pre(self):
        """Test pre condition checking.
        """

        self.assertEqual(self.pre_count, 0)
        self.assertEqual(self.post_count, 0)
        result = self._test()
        self.assertEqual(self.pre_count, 1)
        self.assertEqual(self.post_count, 1)
        self.assertEqual(result, self)


class AnnotationCheckerTest(UTCase):
    """Test AnnotationChecker.
    """

    class TestAnnotationChecker(AnnotationChecker):

        def __init__(self, utcase, *args, **kwargs):

            super(AnnotationCheckerTest.TestAnnotationChecker, self).__init__(
                *args, **kwargs
            )
            self.utcase = utcase

        def _interception(self, joinpoint, *args, **kwargs):

            self.utcase.count += 1

            return joinpoint.proceed()

    def setUp(self):

        self.count = 0
        self.annotation = AnnotationCheckerTest.TestAnnotationChecker(self)

    def tearDown(self):

        self.annotation.__del__()
        del self.annotation

    def test_annotation_class(self):
        """Test to annotate an annotation class.
        """

        self.annotation(Annotation)
        annotation = Annotation()
        annotation(Annotation)
        annotation.__del__()
        self.assertEqual(self.count, 1)

    def test_not_Annotation(self):
        """Test to annotate an object which is not an annotation.
        """

        self.assertRaises(Target.Error, self.annotation, self)


class MaxCountTest(UTCase):
    """Test MaxCount annotation.
    """

    def _assertMaxCount(self, count=MaxCount.DEFAULT_COUNT):

        @MaxCount(count)
        class Test(Annotation):
            pass

        # weave count time test on None
        for i in range(count):
            Test()(None)

        # check if next weaving raise an Exception
        self.assertRaises(MaxCount.Error, Test(), None)

    def test_default(self):
        """Test default count.
        """

        self._assertMaxCount()

    def test_one(self):
        """Test one count.
        """

        self._assertMaxCount(1)

    def test_more_than_one(self):
        """Test > 1 count.
        """

        self._assertMaxCount(randint(2, 5))


class TargetTest(UTCase):
    """Test Target Annotation.
    """

    def test_class(self):
        """Test to use type such as types.
        """

        @Target(type)
        class Test(Annotation):
            pass

        # check to weave on a class
        @Test()
        class TestClassBis(object):
            pass

        # check to weave on a namespace
        @Test()
        class TestNSBis():
            pass

        # check to fail on a lambda expression
        self.assertRaises(Target.Error, Test(), lambda x: None)

    def test_callable(self):
        """Test the type callable.
        """

        @Target(callable)
        class Test(Annotation):
            pass

        @Test()
        def test():
            pass

        @Test()
        class TestClass:
            def __call__(self):
                pass

        testInstance = TestClass()

        Test()(testInstance)

        del TestClass.__call__
        testInstance = TestClass()

        self.assertRaises(Target.Error, Test(), testInstance)

    def test_function(self):
        """Test the type function.
        """

        @Target(Target.FUNC)
        class Test(Annotation):
            pass

        @Test()
        def test():
            pass

        class TestClass:
            pass

        self.assertRaises(Target.Error, Test(), TestClass)

    def test_multitypes(self):
        """Test multi types.
        """

        class TestA:
            pass

        class TestB:
            pass

        class TestAB(TestA, TestB):
            pass

        @Target([TestA, TestB])
        class Test(Annotation):
            pass

        # check types
        Test()(TestA)
        Test()(TestB)
        Test()(TestAB)
        # check instances is not checked
        self.assertRaises(Target.Error, Test(), TestA())
        self.assertRaises(Target.Error, Test(), TestB())
        self.assertRaises(Target.Error, Test(), TestAB())

        @Target([TestA, TestB], instances=True)
        class Test(Annotation):
            pass

        # check types
        Test()(TestA)
        Test()(TestB)
        Test()(TestAB)
        # check instances
        Test()(TestA())
        Test()(TestB())
        Test()(TestAB())

        @Target([TestA, TestB], rule=Target.AND)
        class Test(Annotation):
            pass

        # check types
        self.assertRaises(Target.Error, Test(), TestA)
        self.assertRaises(Target.Error, Test(), TestB)
        Test()(TestAB)
        # check instances is not checked
        self.assertRaises(Target.Error, Test(), TestA())
        self.assertRaises(Target.Error, Test(), TestB())
        self.assertRaises(Target.Error, Test(), TestAB())

        @Target([TestA, TestB], rule=Target.AND, instances=True)
        class Test(Annotation):
            pass

        # check types
        self.assertRaises(Target.Error, Test(), TestA)
        self.assertRaises(Target.Error, Test(), TestB)
        Test()(TestAB)
        # check instances is checked
        self.assertRaises(Target.Error, Test(), TestA())
        self.assertRaises(Target.Error, Test(), TestB())
        Test()(TestAB())


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase

from ..core import Annotation
from ..oop import Transform, Mixin, MethodMixin, Deprecated, Singleton


class TransformTest(UTCase):
    """Test Transform Annotation.
    """

    def test_default(self):

        @Transform()
        class A:
            pass

        self.assertTrue(issubclass(A, Annotation))


class OOPTests(UTCase):

    def setUp(self):
        pass

    def testMixIn(self):

        class A(object):
            def plop(self):
                pass

        plop = A.plop

        mixedins_by_name = Mixin.get_mixedins_by_name(A)

        self.assertEqual(len(mixedins_by_name), 0)

        Mixin.set_mixin(A, 2, 'a')
        Mixin.set_mixin(A, lambda: None, 'b')
        Mixin.set_mixin(A, None, 'plop')
        Mixin.set_mixin(A, None, 'plop')
        Mixin.set_mixin(A, None, 'plop')

        mixedins_by_name = Mixin.get_mixedins_by_name(A)

        self.assertEqual(len(mixedins_by_name), 3)
        self.assertEqual(A.a, 2)
        self.assertIsNone(A.plop)

        Mixin.remove_mixin(A, 'plop')

        mixedins_by_name = Mixin.get_mixedins_by_name(A)

        self.assertEqual(len(mixedins_by_name), 3)

        Mixin.remove_mixins(A)

        mixedins_by_name = Mixin.get_mixedins_by_name(A)

        self.assertEqual(len(mixedins_by_name), 0)
        self.assertFalse(hasattr(A, 'a'))
        self.assertEqual(plop, A.plop)

    def testClassMixIn(self):

        class ClassForMixIn(object):
            def get_1(self):
                return 1

            def get_2(self):
                return 2

        def get_3(self):
            return self.a

        a = None

        @Mixin(ClassForMixIn, lambda: None, get_3=get_3, a=a)
        class MixedInClass(object):
            def get_1(self):
                return '1'

            def get_3(self):
                return 3

        self.assertTrue(hasattr(MixedInClass, 'get_2'))
        self.assertTrue(hasattr(MixedInClass, 'a'))
        self.assertTrue(hasattr(MixedInClass, '<lambda>'))

        mixedInstance = MixedInClass()

        self.assertEqual(mixedInstance.a, None)
        self.assertEqual(mixedInstance.get_3(), a)
        self.assertIs(mixedInstance.get_1(), 1)

        Mixin.remove_mixins(MixedInClass)
        self.assertEqual(mixedInstance.get_3(), 3)
        self.assertIs(mixedInstance.get_1(), '1')

        Mixin(ClassForMixIn, lambda: None, get_3=get_3, a=a)(MixedInClass)

        self.assertTrue(hasattr(MixedInClass, 'get_2'))
        self.assertTrue(hasattr(MixedInClass, 'a'))
        self.assertTrue(hasattr(MixedInClass, '<lambda>'))

        self.assertIsNone(mixedInstance.a)
        self.assertEqual(mixedInstance.get_3(), a)
        self.assertIs(mixedInstance.get_1(), 1)

    def testMethodMixIn(self):

        def test(self):
            return None

        class MixedInClass(object):
            @MethodMixin(test)
            def get_1(self):
                return 1

        self.assertTrue(hasattr(MixedInClass, 'get_1'))

        mixedInstance = MixedInClass()

        self.assertIsNone(mixedInstance.get_1())

        class MixedInClass(object):
            def get_1(self):
                return 1

        MethodMixin(test)(MixedInClass.get_1, ctx=MixedInClass)

        self.assertTrue(hasattr(MixedInClass, 'get_1'))

        mixedInstance = MixedInClass()

        self.assertIsNone(mixedInstance.get_1())

        Mixin.remove_mixins(MixedInClass)

        self.assertEqual(mixedInstance.get_1(), 1)

    def testDeprecated(self):

        @Deprecated()
        def b():
            pass

        b()

    def testSingleton(self):

        @Singleton(a=1)
        class A(object):
            def __init__(self, a):
                self.a = a

            def get_a(self):
                return self.a

        self.assertEqual(A, A())

        self.assertEqual(A(), A())

        self.assertEqual(A.get_a(), A().get_a())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Annotations dedicated to object oriented programming."""

from six import PY2, get_method_function, get_function_code

from .core import Annotation
from .interception import PrivateInterceptor

from types import MethodType

from inspect import getmembers, isfunction, ismethod, isclass

from warnings import warn_explicit

__all__ = ['Transform', 'Mixin', 'Deprecated', 'Singleton', 'MethodMixin']


class Transform(Annotation):
    """Transform a class into an annotation or something else if parameters are
    different.
    """

    NAME = 'name'
    BASES = 'bases'
    DICT = 'dict'
    UPDATE = 'update'

    __slots__ = (NAME, BASES, DICT, UPDATE) + Annotation.__slots__

    def __init__(
            self, name=None, bases=None, _dict=None, update=True,
            *args, **kwargs
    ):
        """According to type constructor parameters, you can specifiy name,
        bases and dict.

        :param str name: new class name.
        :param tuple bases: new class bases.
        :param dict dict: class __dict__.
        :param bool update: if True, update new properties on the new class.
            Otherwise replace old ones.
        """

        super(Transform, self).__init__(*args, **kwargs)

        self.name = name
        self.bases = (Annotation,) if bases is None else bases
        self.dict = {} if _dict is None else _dict
        self.update = update

    def _bind_target(self, target, ctx=None):

        result = target
        # do something only for class objects
        if isclass(target):

            name = target.__name__ if self.name is None else self.name
            # if update is True, update target properties
            bases = tuple(set((target,) + self.bases))

            if self.update:
                _dict = target.__dict__.copy()
                _dict.update(self.dict)

            else:  # else use new properties
                _dict = self.dict

            # get the new type related to name, bases and dict
            result = type(name, bases, _dict)

        else:
            raise TypeError(
                'Wrong annotated element type {0}, class expected.'
                .format(target)
            )

        return result


class Mixin(Annotation):
    """Annotation which enrichs a target with Mixin.

    For every defined mixin, a private couple of (name, array of mixed items)
    is created into the target in order to go back in a no mixin state.
    """

    class MixInError(Exception):
        """Raised for any Mixin error."""

    #: key to organize a dictionary of mixed elements by name in targets.
    __MIXEDIN_KEY__ = '__MIXEDIN__'

    #: Key to identify if a mixin add or replace a content.
    __NEW_CONTENT_KEY__ = '__NEW_CONTENT__'

    def __init__(self, classes=(), *attributes, **named_attributes):

        super(Mixin, self).__init__()

        self.classes = classes if isinstance(classes, tuple) else (classes,)
        self.attributes = attributes
        self.named_attributes = named_attributes

    def on_bind_target(self, target, ctx=None):

        super(Mixin, self).on_bind_target(target)

        for cls in self.classes:
            Mixin.mixin(target, cls)

        for attribute in self.attributes:
            Mixin.mixin(target, attribute)

        for name in self.named_attributes:
            attribute = self.named_attributes[name]
            Mixin.mixin(target, attribute, name)

    @staticmethod
    def mixin_class(target, cls):
        """Mix cls content in target."""

        for name, field in getmembers(cls):
            Mixin.mixin(target, field, name)

    @staticmethod
    def mixin_function_or_method(target, routine, name=None, isbound=False):
        """Mixin a routine into the target.

        :param routine: routine to mix in target.
        :param str name: mixin name. Routine name by default.
        :param bool isbound: If True (False by default), the mixin result is a
            bound method to target.
        """

        function = None

        if isfunction(routine):
            function = routine

        elif ismethod(routine):
            function = get_method_function(routine)

        else:
            raise Mixin.MixInError(
                "{0} must be a function or a method.".format(routine))

        if name is None:
            name = routine.__name__

        if not isclass(target) or isbound:
            _type = type(target)
            method_args = [function, target]

            if PY2:
                method_args += _type

            result = MethodType(*method_args)

        else:

            if PY2:
                result = MethodType(function, None, target)

            else:
                result = function

        Mixin.set_mixin(target, result, name)

        return result

    @staticmethod
    def mixin(target, resource, name=None):
        """Do the correct mixin depending on the type of input resource.

        - Method or Function: mixin_function_or_method.
        - class: mixin_class.
        - other: set_mixin.

        And returns the result of the choosen method (one or a list of mixins).
        """

        result = None

        if ismethod(resource) or isfunction(resource):
            result = Mixin.mixin_function_or_method(target, resource, name)

        elif isclass(resource):
            result = list()

            for name, content in getmembers(resource):
                if isclass(content):
                    mixin = Mixin.set_mixin(target, content, name)

                else:
                    mixin = Mixin.mixin(target, content, name)

                result.append(mixin)

        else:
            result = Mixin.set_mixin(target, resource, name)

        return result

    @staticmethod
    def get_mixedins_by_name(target):
        """Get a set of couple (name, field) of target mixedin."""

        result = getattr(target, Mixin.__MIXEDIN_KEY__, None)

        if result is None:
            result = dict()
            setattr(target, Mixin.__MIXEDIN_KEY__, result)

        return result

    @staticmethod
    def set_mixin(target, resource, name=None, override=True):
        """Set a resource and returns the mixed one in target content or
        Mixin.__NEW_CONTENT_KEY__ if resource name didn't exist in target.

        :param str name: target content item to mix with the resource.
        :param bool override: If True (default) permits to replace an old
            resource by the new one.
        """

        if name is None and not hasattr(resource, '__name__'):
            raise Mixin.MixInError(
                "name must be given or resource {0} can't be anonymous"
                .format(resource))

        result = None

        name = resource.__name__ if name is None else name

        mixedins_by_name = Mixin.get_mixedins_by_name(target)

        result = getattr(target, name, Mixin.__NEW_CONTENT_KEY__)

        if override or result == Mixin.__NEW_CONTENT_KEY__:

            if name not in mixedins_by_name:
                mixedins_by_name[name] = (result,)
            else:
                mixedins_by_name[name] += (result,)

            try:
                setattr(target, name, resource)

            except (AttributeError, TypeError):

                if len(mixedins_by_name[name]) == 1:
                    del mixedins_by_name[name]
                    if len(mixedins_by_name) == 0:
                        delattr(target, Mixin.__MIXEDIN_KEY__)

                else:
                    mixedins_by_name[name] = mixedins_by_name[name][:-2]

                result = None
        else:
            result = None

        return result

    @staticmethod
    def remove_mixin(target, name, mixedin=None, replace=True):
        """Remove a mixin with name (and reference) from targetand returns the
        replaced one or None.

        :param mixedin: a mixedin value or the last defined mixedin if is None
            (by default).
        :param bool replace: If True (default), the removed mixedin replaces
            the current mixin.
        """

        try:
            result = getattr(target, name)

        except AttributeError:
            raise Mixin.MixInError(
                "No mixin {0} exists in {1}".format(name, target)
            )

        mixedins_by_name = Mixin.get_mixedins_by_name(target)

        mixedins = mixedins_by_name.get(name)

        if mixedins:

            if mixedin is None:
                mixedin = mixedins[-1]
                mixedins = mixedins[:-2]

            else:

                try:
                    index = mixedins.index(mixedin)

                except ValueError:
                    raise Mixin.MixInError(
                        "Mixedin {0} with name {1} does not exist \
                        in target {2}"
                        .format(mixedin, name, target)
                    )

                mixedins = mixedins[0:index] + mixedins[index + 1:]

            if len(mixedins) == 0:
                # force to replace/delete the mixin even if replace is False
                # in order to stay in a consistent state
                if mixedin != Mixin.__NEW_CONTENT_KEY__:
                    setattr(target, name, mixedin)

                else:
                    delattr(target, name)

                del mixedins_by_name[name]

            else:
                if replace:
                    setattr(target, name, mixedin)

                mixedins_by_name[name] = mixedins

        else:
            # shouldn't be raised except if removing has been done
            # manually
            raise Mixin.MixInError(
                "No mixin {0} exists in {1}".format(name, target))

        # clean mixedins if no one exists
        if len(mixedins_by_name) == 0:
            delattr(target, Mixin.__MIXEDIN_KEY__)

        return result

    @staticmethod
    def remove_mixins(target):
        """Tries to get back target in a no mixin consistent state.
        """

        mixedins_by_name = Mixin.get_mixedins_by_name(target).copy()

        for _name in mixedins_by_name:  # for all named mixins
            while True:  # remove all mixins named _name
                try:
                    Mixin.remove_mixin(target, _name)
                except Mixin.MixInError:
                    break  # leave the loop 'while'


class MethodMixin(Annotation):
    """Apply a mixin on a method."""

    def __init__(self, function, *args, **kwargs):

        super(MethodMixin, self).__init__(*args, **kwargs)

        self.function = function

    def _bind_target(self, target, ctx, *args, **kwargs):

        result = super(MethodMixin, self)._bind_target(
            target=target, ctx=ctx, *args, **kwargs
        )

        if ismethod(result):
            cls = target.im_class
            name = target.__name__
            Mixin.mixin_function_or_method(cls, self.function, name)
            result = target

        else:

            if ctx is not None:
                cls = ctx
                name = target.__name__
                Mixin.mixin_function_or_method(cls, self.function, name)

            result = self.function

        return result


class Deprecated(PrivateInterceptor):
    """Decorator which can be used to mark functions as deprecated. It will
    result in a warning being emitted when the function is used.
    """

    def _interception(self, joinpoint):

        target = joinpoint.target
        target_code = get_function_code(target)
        warn_explicit(
            "Call to deprecated function {0}.".format(target.__name__),
            category=DeprecationWarning,
            filename=target_code.co_filename,
            lineno=target_code.co_firstlineno + 1
        )
        result = joinpoint.proceed()
        return result


class Singleton(Annotation):
    """Transforms cls into a singleton.

    Reference to cls, or to any instance is the same reference.
    """

    def __init__(self, *args, **kwargs):
        super(Singleton, self).__init__()
        self.args = args
        self.kwargs = kwargs

    def _bind_target(self, target, ctx, *args, **kwargs):

        target = super(Singleton, self)._bind_target(
            target, ctx=ctx, *args, **kwargs
        )

        instance = target(*self.args, **self.kwargs)
        instance.__call__ = lambda x=None: instance
        target.__call__ = lambda x=None: instance

        return instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from time import sleep

from inspect import getmembers

from b3j0f.utils.ut import UTCase

from six.moves import range

from ..core import Annotation, StopPropagation, RoutineAnnotation


class TestAnnotation(Annotation):
    """Annotation for inheritance tests."""


class AnnotationTest(UTCase):
    """UT class which creates an annotation and delete it at the end."""

    def setUp(self):
        """Create a new annotation."""

        self.annotation = Annotation()

    def tearDown(self):
        """Delete self.annotation."""

        self.annotation.__del__()
        del self.annotation


class MemoryTest(AnnotationTest):
    """Test InMemory Annotations."""

    def setUp(self):

        super(MemoryTest, self).setUp()

        Annotation.free_memory()

    def tearDown(self):

        super(MemoryTest, self).tearDown()

        Annotation.free_memory()

    def test_not_in_memory(self):
        """Test annotation not in memory
        """

        annotations = Annotation.get_memory_annotations()
        self.assertFalse(annotations)

    def test_in_memory(self):
        """Test if annotation is in memory
        """

        self.annotation = Annotation(in_memory=True)
        annotations = Annotation.get_memory_annotations()
        self.assertEqual(annotations, set((self.annotation,)))

    def test_set_in_memory(self):
        """Test set in_memory
        """

        self.annotation.in_memory = True
        annotations = Annotation.get_memory_annotations()
        self.assertEqual(annotations, set((self.annotation,)))

    def test_set_not_in_memory(self):
        """Test set not in_memory
        """

        self.annotation.in_memory = True
        annotations = Annotation.get_memory_annotations()
        self.assertEqual(annotations, set((self.annotation,)))
        self.annotation.in_memory = False
        annotations = Annotation.get_memory_annotations()
        self.assertFalse(annotations)

    def test_get(self):

        testAnnotation = TestAnnotation(in_memory=True)
        self.annotation.in_memory = True
        annotations = Annotation.get_memory_annotations()
        self.assertEqual(annotations, set((testAnnotation, self.annotation)))
        testAnnotation.__del__()

    def test_get_inheritance(self):

        testAnnotation = TestAnnotation(in_memory=True)
        self.annotation.in_memory = True
        annotations = TestAnnotation.get_memory_annotations()
        self.assertEqual(annotations, set((testAnnotation,)))
        testAnnotation.__del__()

    def test_exclude(self):

        testAnnotation = TestAnnotation(in_memory=True)
        self.annotation.in_memory = True
        annotations = Annotation.get_memory_annotations(exclude=TestAnnotation)
        self.assertEqual(annotations, set((self.annotation,)))
        annotations = Annotation.get_memory_annotations(exclude=Annotation)
        self.assertFalse(annotations)
        testAnnotation.__del__()


class DeleteTest(AnnotationTest):
    """Test annotation deletion."""

    def test_one_annotation(self):
        """Test to delete one bound annotation."""

        self.annotation(self)

        annotations = Annotation.get_annotations(self)

        self.assertEqual(len(annotations), 1)

        self.annotation.__del__()

        annotations = Annotation.get_annotations(self)

        self.assertFalse(annotations)

    def test_two_annotations(self):
        """Test to delete one annotation bound twice on the same element."""

        self.annotation(self)
        self.annotation(self)

        annotations = Annotation.get_annotations(self)

        self.assertEqual(len(annotations), 2)

        self.annotation.__del__()

        annotations = Annotation.get_annotations(self)

        self.assertFalse(annotations)

    def test_two_annotations_wit_two_objects(self):
        """Test to delete one annotations bound to two elements."""

        self.annotation(self)
        self.annotation(DeleteTest)

        annotations = Annotation.get_annotations(self)

        self.assertEqual(len(annotations), 2)

        annotations = Annotation.get_annotations(DeleteTest)

        self.assertEqual(len(annotations), 1)

        self.annotation.__del__()

        annotations = Annotation.get_annotations(self)

        self.assertFalse(annotations)

        annotations = Annotation.get_annotations(DeleteTest)

        self.assertFalse(annotations)


class RemoveTest(AnnotationTest):
    """Test remove class method."""

    def setUp(self):

        super(RemoveTest, self).setUp()

        self.test_annotation = TestAnnotation()

        self.annotation(self)
        self.test_annotation(self)

    def tearDown(self):

        del self.annotation
        del self.test_annotation

    def test(self):
        """Test simple remove."""

        Annotation.remove(self)

        annotations = Annotation.get_annotations(self)

        self.assertFalse(annotations)

    def test_inheritance(self):

        TestAnnotation.remove(self)

        annotations = Annotation.get_annotations(self)

        self.assertEqual(annotations[0], self.annotation)

    def test_exclude(self):

        Annotation.remove(self, exclude=TestAnnotation)

        annotations = Annotation.get_annotations(self)

        self.assertEqual(annotations[0], self.test_annotation)


class OnBindTargetTest(AnnotationTest):
    """Test on_bind_target handler."""

    def setUp(self):

        super(OnBindTargetTest, self).setUp()

        self.count = 0
        self.annotation._on_bind_target = self.on_bind_target

    def on_bind_target(self, annotation, target, ctx, *args, **kwargs):
        self.count += 1

    def test_one(self):

        self.annotation(self)

        self.assertEqual(self.count, 1)

    def test_many(self):

        self.annotation(self)
        self.annotation(self)

        self.assertEqual(self.count, 2)

    def test_many_elts(self):

        self.annotation(self)
        self.annotation(OnBindTargetTest)
        self.annotation(OnBindTargetTest)

        self.assertEqual(self.count, 3)


class TargetsTest(AnnotationTest):
    """Test targets attribute."""

    def test_none(self):

        self.assertFalse(self.annotation.targets)

    def test_one(self):

        self.annotation(self)

        self.assertIn(self, self.annotation.targets)

    def test_many(self):

        self.annotation(self)
        self.annotation(self)

        self.assertIn(self, self.annotation.targets)

    def test_many_many(self):

        self.annotation(self)
        self.annotation(TargetsTest)

        self.assertIn(self, self.annotation.targets)
        self.assertIn(TargetsTest, self.annotation.targets)


class TTLTest(AnnotationTest):
    """Test ttl."""

    def setUp(self):

        super(TTLTest, self).setUp()

        self.ttl = 0.1

    def test_def(self):
        """Test ttl at definition."""

        self.annotation = Annotation(ttl=self.ttl)
        self.annotation(self)

        annotations = Annotation.get_annotations(self)

        self.assertTrue(annotations)

        sleep(2 * self.ttl)

        annotations = Annotation.get_annotations(self)

        self.assertFalse(annotations)

    def test_run(self):
        """Test to set ttl at runtime."""

        self.annotation(self)

        annotations = Annotation.get_annotations(self)

        self.assertTrue(annotations)

        self.annotation.ttl = self.ttl

        sleep(2 * self.ttl)

        annotations = Annotation.get_annotations(self)

        self.assertFalse(annotations)

    def test_run_run(self):
        """Test to change ttl after changing it a first time."""

        self.annotation(self)

        annotations = Annotation.get_annotations(self)

        self.assertTrue(annotations)

        self.annotation.ttl = 5

        self.assertLess(self.annotation.ttl, 5)

        self.annotation.ttl = 10

        self.assertGreater(self.annotation.ttl, 5)

        self.annotation.ttl = None

        self.assertIsNone(self.annotation.ttl)


class GetAnnotationsTest(AnnotationTest):
    """Test to annotate elements."""

    def test_None(self):
        """Test to annotate None."""

        self.annotation(None)

        annotations = Annotation.get_annotations(None)

        self.assertTrue(annotations)

    def test_number(self):
        """Test to annotate a number."""

        self.annotation(1)

        annotations = Annotation.get_annotations(1)

        self.assertTrue(annotations)

    def test_function(self):
        """Test to annotate a function."""

        def test():
            """test function."""

        self.annotation(test)

        annotations = Annotation.get_annotations(test)

        self.assertTrue(annotations)

    def test_builtin(self):
        """Test to annotate a builtin element."""

        self.annotation(range)

        annotations = Annotation.get_annotations(range)

        self.assertTrue(annotations)

    def test_class(self):
        """Test to annotate a class."""

        class Test(object):
            """Test class."""

        self.annotation(Test)

        annotations = Annotation.get_annotations(Test)

        self.assertTrue(annotations)

    def test_dclass(self):
        """Test to annotate directly a class."""

        @Annotation()
        class Test(object):
            """Test class."""

        annotations = Annotation.get_annotations(Test)

        self.assertTrue(annotations)

    def test_namespace(self):
        """Test to annotate a namespace."""

        class Test:
            """Test namespace."""

        self.annotation(Test)

        annotations = Annotation.get_annotations(Test)

        self.assertTrue(annotations)

    def test_method(self):
        """Test to annotate a method."""

        class Test:
            """Test class."""
            def test(self):
                """Test method."""

        self.annotation(Test.test, ctx=Test)

        annotations = Annotation.get_annotations(Test.test, ctx=Test)

        self.assertTrue(annotations)

    def test_dmethod(self):
        """Test to annotate directly a method."""

        class Test:
            """Test class."""
            @Annotation()
            def test(self):
                """Test method."""

        annotations = Annotation.get_annotations(Test.test, ctx=Test)

        self.assertTrue(annotations)

    def test_boundmethod(self):
        """Test to annotate a bound method."""

        class Test:
            """Test class."""
            def test(self):
                """Test method."""

        test = Test()

        self.annotation(test.test)

        annotations = Annotation.get_annotations(test.test)

        self.assertTrue(annotations)

    def test_instance(self):
        """Test to annotate an instance."""

        class Test:
            """Test class."""

        test = Test()

        self.annotation(test)

        annotations = Annotation.get_annotations(test)

        self.assertTrue(annotations)

    def test_module(self):
        """Test to annotate a module."""

        import sys

        self.annotation(sys)

        annotations = Annotation.get_annotations(sys)

        self.assertTrue(annotations)

    def test_dconstructor(self):
        """Test to annotate directly a constructor."""

        class Test(object):

            @Annotation()
            def __init__(self):
                pass

        #self.annotation(Test.__init__, ctx=Test)

        annotations = Annotation.get_annotations(Test.__init__, ctx=Test)

        self.assertTrue(annotations)

    def test_constructor(self):
        """Test to annotate a constructor."""

        class Test(object):

            def __init__(self):
                pass

        self.annotation(Test.__init__, ctx=Test)

        annotations = Annotation.get_annotations(Test.__init__, ctx=Test)

        self.assertTrue(annotations)

    def test_boundconstructor(self):
        """Test to annotate a bound constructor."""

        class Test(object):

            def __init__(self):
                pass

        test = Test()

        self.annotation(test.__init__, ctx=test)

        annotations = Annotation.get_annotations(test.__init__, ctx=test)

        self.assertTrue(annotations)

    def test_dboundconstructor(self):
        """Test to annotate directly a bound constructor."""

        class Test(object):
            @Annotation()
            def __init__(self):
                pass

        test = Test()

        annotations = Annotation.get_annotations(test.__init__, ctx=test)

        self.assertTrue(annotations)


class GetLocalAnnotationsTest(AnnotationTest):
    """Test get local annotatations."""

    def test_None(self):
        """Test to annotate None.
        """

        self.annotation(None)

        annotations = Annotation.get_local_annotations(None)

        self.assertTrue(annotations)

    def test_number(self):
        """Test to annotate a number.
        """

        self.annotation(1)

        annotations = Annotation.get_local_annotations(1)

        self.assertTrue(annotations)

    def test_function(self):
        """Test to annotate a function.
        """

        def test():
            pass

        self.annotation(test)

        annotations = Annotation.get_local_annotations(test)

        self.assertTrue(annotations)

    def test_builtin(self):
        """Test to annotate a builtin element.
        """

        self.annotation(range)

        annotations = Annotation.get_local_annotations(range)

        self.assertTrue(annotations)

    def test_class(self):
        """Test to annotate a class.
        """

        class Test(object):
            pass

        self.annotation(Test)

        annotations = Annotation.get_local_annotations(Test)

        self.assertTrue(annotations)

    def test_namespace(self):
        """Test to annotate a namespace.
        """

        class Test:
            pass

        self.annotation(Test)

        annotations = Annotation.get_local_annotations(Test)

        self.assertTrue(annotations)

    def test_method(self):
        """Test to annotate a method.
        """

        class Test:
            def test(self):
                pass

        self.annotation(Test.test)

        annotations = Annotation.get_local_annotations(Test.test)

        self.assertTrue(annotations)

    def test_boundmethod(self):
        """Test to annotate a bound method.
        """

        class Test:
            def test(self):
                pass

        test = Test()

        self.annotation(test.test)

        annotations = Annotation.get_local_annotations(test.test)

        self.assertTrue(annotations)

    def test_instance(self):
        """Test to annotate an instance.
        """

        class Test:
                pass

        test = Test()

        self.annotation(test)

        annotations = Annotation.get_local_annotations(test)

        self.assertTrue(annotations)

    def test_module(self):
        """Test to annotate a module.
        """

        import sys

        self.annotation(sys)

        annotations = Annotation.get_local_annotations(sys)

        self.assertTrue(annotations)

    def test_dconstructor(self):
        """Test to annotate directly a constructor.
        """

        class Test:
            @Annotation()
            def __init__(self):
                pass

        annotations = Annotation.get_local_annotations(Test.__init__)

        self.assertTrue(annotations)

    def test_constructor(self):
        """Test to annotate a constructor.
        """

        class Test:

            def __init__(self):
                pass

        self.annotation(Test.__init__)

        annotations = Annotation.get_local_annotations(Test.__init__)

        self.assertTrue(annotations)

    def test_boundconstructor(self):
        """Test to annotate a bound constructor.
        """

        class Test:
            def __init__(self):
                pass

        test = Test()

        self.annotation(test.__init__, ctx=test)

        annotations = Annotation.get_local_annotations(test.__init__, ctx=test)

        self.assertTrue(annotations)

    def test_dboundconstructor(self):
        """Test to annotate directly a bound constructor.
        """

        class Test:
            @Annotation()
            def __init__(self):
                pass

        test = Test()

        annotations = Annotation.get_local_annotations(test.__init__, ctx=test)

        self.assertTrue(annotations)


class GetParameterizedAnnotationsTest(AnnotationTest):

    def setUp(self):

        super(GetParameterizedAnnotationsTest, self).setUp()

        class BaseTest:
            pass

        class Test(BaseTest):
            pass

        self.annotation(BaseTest)
        self.annotation(Test)

        self.Test = Test
        self.BaseTest = BaseTest

        annotations = Annotation.get_annotations(BaseTest)

        self.assertEqual(len(annotations), 1)

        annotations = Annotation.get_annotations(Test)

        self.assertEqual(len(annotations), 2)

    def test_override(self):
        """
        Test to override annotation
        """

        annotations = Annotation.get_annotations(self.Test)

        self.assertEqual(len(annotations), 2)

        self.annotation.override = True

        annotations = Annotation.get_annotations(self.Test)

        self.assertEqual(len(annotations), 0)

    def test_propagate(self):
        """
        Test to propagate annotation
        """

        self.annotation.propagate = False

        annotations = Annotation.get_annotations(self.Test)

        self.assertEqual(len(annotations), 1)

    def test_exclude(self):
        """
        Test to exclude annotations
        """

        test_annotation = TestAnnotation()

        test_annotation(self.Test)
        test_annotation(self.BaseTest)

        annotations = Annotation.get_annotations(self.Test)

        self.assertEqual(len(annotations), 4)

        annotations = Annotation.get_annotations(
            self.Test, exclude=TestAnnotation)

        self.assertEqual(len(annotations), 2)

    def test_stop_propagation(self):
        """
        Test Stop propagation annotation
        """

        stop_propagation = StopPropagation(Annotation)

        stop_propagation(self.BaseTest)

        annotations = Annotation.get_annotations(self.Test)

        self.assertEqual(len(annotations), 1)

        stop_propagation.__del__()

        annotations = Annotation.get_annotations(self.Test)

        self.assertEqual(len(annotations), 2)

        stop_propagation(self.Test)

        annotations = Annotation.get_annotations(self.Test)

        self.assertEqual(len(annotations), 0)


class GetLocalParameterizedAnnotationsTest(AnnotationTest):

    def setUp(self):

        super(GetLocalParameterizedAnnotationsTest, self).setUp()

        class BaseTest:
            pass

        class Test(BaseTest):
            pass

        self.annotation(BaseTest)
        self.annotation(Test)

        self.Test = Test
        self.BaseTest = BaseTest

        annotations = Annotation.get_local_annotations(BaseTest)

        self.assertEqual(len(annotations), 1)

        annotations = Annotation.get_local_annotations(Test)

        self.assertEqual(len(annotations), 1)

    def test_override(self):
        """
        Test to override annotation
        """

        annotations = Annotation.get_local_annotations(self.Test)

        self.assertEqual(len(annotations), 1)

        self.annotation.override = True

        annotations = Annotation.get_local_annotations(self.Test)

        self.assertEqual(len(annotations), 1)

    def test_propagate(self):
        """
        Test to propagate annotation
        """

        self.annotation.propagate = False

        annotations = Annotation.get_local_annotations(self.Test)

        self.assertEqual(len(annotations), 1)

    def test_exclude(self):
        """
        Test to exclude annotations
        """

        test_annotation = TestAnnotation()

        test_annotation(self.Test)
        test_annotation(self.BaseTest)

        annotations = Annotation.get_local_annotations(self.Test)

        self.assertEqual(len(annotations), 2)

        annotations = Annotation.get_local_annotations(
            self.Test, exclude=TestAnnotation)

        self.assertEqual(len(annotations), 1)

    def test_stop_propagation(self):
        """
        Test Stop propagation annotation
        """

        stop_propagation = StopPropagation(Annotation)

        stop_propagation(self.BaseTest)

        annotations = Annotation.get_local_annotations(self.Test)

        self.assertEqual(len(annotations), 1)

        stop_propagation.__del__()

        annotations = Annotation.get_local_annotations(self.Test)

        self.assertEqual(len(annotations), 1)

        stop_propagation(self.Test)

        annotations = Annotation.get_local_annotations(self.Test)

        self.assertEqual(len(annotations), 2)


class GetAnnotatedFields(AnnotationTest):
    """
    Test get_annotated_fields class method
    """

    def test_class(self):

        cls = object

        members = set()

        for _, member in getmembers(cls):

            try:
                self.annotation(member, ctx=cls)
            except:
                pass
            else:
                members.add(member)

        annotated_fields = Annotation.get_annotated_fields(object)

        self.assertEqual(len(annotated_fields), len(members))

        for annotated_field in annotated_fields:

            annotations = annotated_fields[annotated_field]

            self.assertIs(annotations[0], self.annotation)


class RoutineAnnotationTest(AnnotationTest):

    def test(self):

        self.routine = 'routine'
        self.params = 'params'
        self.result = 'result'

        self.annotation = RoutineAnnotation(
            routine=self.routine, params=self.params, result=self.result
        )

        self.assertEqual(self.annotation.routine, self.routine)
        self.assertEqual(self.annotation.params, self.params)
        self.assertEqual(self.annotation.result, self.result)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

from unittest import main

from b3j0f.utils.ut import UTCase

from ..interception import (
    Interceptor, PrivateInterceptor, PrivateCallInterceptor, CallInterceptor
)


class InterceptorTest(UTCase):
    """
    Base class for other tests
    """

    def get_interceptor(self):

        raise NotImplementedError()

    def get_target(self):

        raise NotImplementedError()

    def setUp(self):

        self.count = 0

        self.interceptor = self.get_interceptor()
        self.target = self.get_target()
        self.interceptor(self.target)

    def interception(self, joinpoint):

        self.count += 1

        return joinpoint.proceed()

    def tearDown(self):
        """
        Delete self.interceptor
        """

        self.interceptor.__del__()
        del self.interceptor


class InterceptionTest(InterceptorTest):
    """
    Test interception
    """

    def get_interceptor(self):

        return Interceptor(interception=self.interception)

    def get_target(self):

        return lambda: None

    def call_target(self):
        """Call target.
        """

        self.target()

    def test_one_one(self):
        """
        Test one annotation and one call
        """

        self.call_target()

        self.assertEqual(self.count, 1)

    def test_one_two(self):
        """
        Test one annotation and two calls
        """
        self.call_target()
        self.call_target()

        self.assertEqual(self.count, 2)

    def test_two_one(self):
        """
        Test two annotation and one call
        """

        self.interceptor(self.target)

        self.call_target()

        self.assertEqual(self.count, 2)

    def test_two_two(self):
        """
        Test two annotations and two calls
        """

        self.interceptor(self.target)

        self.call_target()
        self.call_target()

        self.assertEqual(self.count, 4)

    def test_enable(self):

        self.interceptor.enable = True

        self.call_target()

        self.assertEqual(self.count, 1)

    def test_disable(self):

        self.interceptor.enable = False

        self.call_target()

        self.assertEqual(self.count, 0)

    def test_enables(self):

        Interceptor.set_enable(self.target, enable=True)

        self.call_target()

        self.assertEqual(self.count, 1)

    def test_disables(self):

        Interceptor.set_enable(self.target, enable=False)

        self.call_target()

        self.assertEqual(self.count, 0)


class PrivateInterceptorTest(InterceptorTest):
    """
    Test interception
    """

    class TestPrivateInterceptor(PrivateInterceptor):

        def __init__(self, utcase):

            super(
                PrivateInterceptorTest.TestPrivateInterceptor, self
            ).__init__()

            self.utcase = utcase

        def _interception(self, joinpoint, *args, **kwargs):

            return self.utcase.interception(joinpoint)

    def get_interceptor(self):

        return PrivateInterceptorTest.TestPrivateInterceptor(self)

    def get_target(self):

        return lambda: None


class CallInterceptorTest(InterceptorTest):
    """
    Test interception
    """
    class Test(object):

        def __call__(self):

            pass

    def get_interceptor(self):

        return CallInterceptor(interception=self.interception)

    def get_target(self):

        return CallInterceptorTest.Test

    def call_target(self):

        self.target()()


class PrivateCallInterceptorTest(InterceptorTest):
    """
    Test interception
    """

    class TestCallInterceptor(PrivateCallInterceptor):

        def __init__(self, utcase):

            super(
                PrivateCallInterceptorTest.TestCallInterceptor,
                self
            ).__init__()
            self.utcase = utcase

        def _interception(self, joinpoint):
            return self.utcase.interception(joinpoint)

    class Test(object):

        def __call__(self):
            pass

    def get_interceptor(self):

        return PrivateCallInterceptorTest.TestCallInterceptor(self)

    def get_target(self):

        return CallInterceptorTest.Test

    def call_target(self):

        self.target()()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Annotation package."""

__all__ = [
    '__version__',  # version
    'Annotation',
    'Synchronized', 'SynchronizedClass',
    'Asynchronous', 'TimeOut', 'Wait', 'Observable',
    'Types', 'types', 'Curried', 'curried', 'Retries',
    'Condition', 'MaxCount', 'Target',
    'Interceptor',
    'PrivateInterceptor', 'CallInterceptor', 'PrivateCallInterceptor',
    'Transform', 'Mixin', 'Deprecated', 'Singleton', 'MethodMixin'
]

from .version import __version__

from .core import Annotation
from .async import (
    Synchronized, SynchronizedClass, Asynchronous, TimeOut, Wait, Observable
)
from .call import Types, types, Curried, curried, Retries
from .check import Condition, MaxCount, Target
from .interception import (
    Interceptor, PrivateInterceptor, CallInterceptor, PrivateCallInterceptor
)
from .oop import (
    Transform, Mixin, Deprecated, Singleton, MethodMixin
)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Interceptors dedicated to decorate decorations."""

from .core import Annotation
from .interception import PrivateInterceptor, PrivateCallInterceptor

from b3j0f.utils.iterable import ensureiterable

from types import FunctionType

from inspect import isclass, isroutine

from six import callable

__all__ = ['Condition', 'MaxCount', 'Target']


class Condition(PrivateInterceptor):
    """Apply a pre/post condition on an annotated element call."""

    class ConditionError(Exception):
        """Handle condition errors."""

    class PreConditionError(ConditionError):
        """Handle pre condition errors."""

    class PostConditionError(ConditionError):
        """Handle post condition errors."""

    #: pre condition attribute name
    PRE_COND = 'pre_cond'

    #: post condition attribute name
    POST_COND = 'post_cond'

    #: result attribute name
    RESULT = 'result'

    __slots__ = (PRE_COND, POST_COND) + PrivateInterceptor.__slots__

    def __init__(self, pre_cond=None, post_cond=None, *args, **kwargs):
        """
        :param pre_cond: function called before target call. Parameters
            are self annotation and AdvicesExecutor.
        :param post_cond: function called after target call. Parameters
            are self annotation, call result and AdvicesExecutor.
        """

        super(Condition, self).__init__(*args, **kwargs)

        self.pre_cond = pre_cond
        self.post_cond = post_cond

    def _interception(self, joinpoint):
        """Intercept call of joinpoint callee in doing pre/post conditions.
        """

        if self.pre_cond is not None:
            self.pre_cond(joinpoint)

        result = joinpoint.proceed()

        if self.post_cond is not None:
            joinpoint.exec_ctx[Condition.RESULT] = result
            self.post_cond(joinpoint)

        return result


class AnnotationChecker(PrivateInterceptor):
    """Annotation dedicated to intercept annotation target binding."""

    __slots = PrivateCallInterceptor.__slots__

    #: bind_target pointcut
    __BIND_TARGET__ = 'bind_target'

    def __init__(self, *args, **kwargs):

        super(AnnotationChecker, self).__init__(
            pointcut=AnnotationChecker.__BIND_TARGET__, *args, **kwargs
        )


class MaxCount(AnnotationChecker):
    """Set a maximum count of target instantiation by annotated element.
    """

    class Error(Exception):
        """Handle MaxCount errors."""

    __COUNT_KEY__ = 'count'

    DEFAULT_COUNT = 1

    __slots__ = (__COUNT_KEY__, ) + AnnotationChecker.__slots__

    def __init__(self, count=DEFAULT_COUNT, *args, **kwargs):
        """
        :param int count: maximal target instanciation by annotated element.
        """

        super(MaxCount, self).__init__(*args, **kwargs)

        self.count = MaxCount.DEFAULT_COUNT if count is None else count

    def _interception(self, joinpoint):

        target = joinpoint.kwargs['target']
        annotation = joinpoint.kwargs['self']

        annotation_class = annotation.__class__
        annotations = annotation_class.get_annotations(target)

        if len(annotations) >= self.count:
            raise MaxCount.Error(
                '{0} calls of {1} on {2}'.format(
                    self.count + 1, annotation, target)
            )

        result = joinpoint.proceed()

        return result

# apply MaxCount on itself
MaxCount()(MaxCount)


@MaxCount()
class Target(AnnotationChecker):
    """Check type of all decorated elements by this decorated Annotation in
    using a list of types.

    This list of types is checked related to logical rules such as OR and AND.

    - OR: true if at least one type is checked.
    - AND: true if all types are checked.
    """

    class Error(Exception):
        """Handle Target errors."""

    FUNC = FunctionType  #: function type.
    CALLABLE = callable  #: callable type.
    ROUTINE = 'routine'  #: routine type.

    OR = 'or'  #: or rule.
    AND = 'and'  #: and rule.

    DEFAULT_RULE = OR  #: default rule
    DEFAULT_INSTANCES = False  #: default instances condition

    TYPES = 'types'
    RULE = 'rule'
    INSTANCES = 'instances'

    __slots__ = (TYPES, RULE, INSTANCES) + PrivateCallInterceptor.__slots__

    def __init__(
            self, types=None, rule=DEFAULT_RULE, instances=False,
            *args, **kwargs
    ):
        """
        :param types: type(s) to check. The function ``callable`` can be used.
        :type types: type or list
        :param str rule: set condition on input types.
        :param bool instances: if True, check types such and instance types.
        """
        super(Target, self).__init__(*args, **kwargs)

        self.types = () if types is None else ensureiterable(types)
        self.rule = rule
        self.instances = instances

    def _interception(self, joinpoint):

        raiseexception = self.rule == Target.OR

        target = joinpoint.kwargs['target']
        annotation = joinpoint.kwargs['self']

        for _type in self.types:
            if (
                    (_type == Target.ROUTINE and isroutine(target))
                    or
                    (_type == Target.FUNC and isinstance(target, Target.FUNC))
                    or
                    (_type == type and isclass(target))
                    or
                    (_type is callable and callable(target))
                    or
                    (
                        _type is not callable
                        and (
                            (isclass(target) and issubclass(target, _type))
                            or (self.instances and isinstance(target, _type))
                        )
                    )
            ):
                if self.rule == Target.OR:
                    raiseexception = False
                    break

            elif self.rule == Target.AND:
                raiseexception = True
                break

        if raiseexception:
            interceptor_type = annotation.__class__

            raise Target.Error(
                "{0} is not allowed by {1}. Must be {2} {3}".format(
                    target,
                    interceptor_type,
                    'among' if self.rule == Target.OR else 'all',
                    self.types)
            )

        result = joinpoint.proceed()

        return result

# ensure AnnotationChecker is dedicated to Annotation
Target(Annotation)(AnnotationChecker)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""b3j0f.annotation building script."""

from setuptools import setup, find_packages

from os.path import abspath, dirname, join

from re import compile as re_compile, S as re_S

NAME = 'b3j0f.annotation'  # library name

_namepath = NAME.replace('.', '/')

BASEPATH = dirname(abspath(__file__))

# get long description from setup directory abspath
with open(join(BASEPATH, 'README.rst')) as f:
    DESC = f.read()

# Get the version - do not use normal import because it does break coverage
# thanks to the python jira project
# (https://github.com/pycontribs/jira/blob/master/setup.py)
with open(join(BASEPATH, _namepath, 'version.py')) as f:
    stream = f.read()
    regex = r'.*__version__ = \'(.*?)\''
    VERSION = re_compile(regex, re_S).match(stream).group(1)

DEPENDENCIES = []
with open(join(BASEPATH, 'requirements.txt')) as f:
    DEPENDENCIES = list(line for line in f.readlines())

KEYWORDS = [
    'annotation', 'decorator', 'type', 'check', 'async', 'interceptor', 'oop',
    'call', 'mixin', 'transform', 'curried', 'retries', 'deprecated',
    'reflect', 'reflectivity'
]

DESCRIPTION = 'Reflective annotation library for python'

URL = 'https://github.com/{0}'.format(_namepath)

setup(
    name=NAME,
    version=VERSION,
    install_requires=DEPENDENCIES,
    packages=find_packages(exclude=['test.*', '*.test.*']),
    author='b3j0f',
    author_email='jlabejof@yahoo.fr',
    description=DESCRIPTION,
    long_description=DESC,
    url=URL,
    license='MIT License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: French',
        'Operating System :: OS Independent',
        'Topic :: Utilities',
        'Topic :: Software Development',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython'
    ],
    test_suite='b3j0f',
    keywords=KEYWORDS
)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""This module defines the Annotation class.

Such Annotation is close to the reflective paradigm in having himself its own
    lifecycle independently from its annotated elements (called commonly
        target).

Such Annotation:

- can annotate all python object, from classes to class instances.
- have their own business behaviour.
- are reusable to annotate several objects.
- can propagate their annotation capability to class inheritance tree.

.. limitations:
    It is impossible to annotate None methods.
"""

__all__ = ['Annotation', 'StopPropagation', 'RoutineAnnotation']

from b3j0f.utils.property import (
    put_properties, del_properties, get_local_property, get_property
)

from six import get_method_function

from time import time

try:
    from threading import Timer
except ImportError:
    from dummy_threading import Timer

from inspect import ismethod, getmembers, isfunction


class Annotation(object):
    """Base class for all annotations defined in this library.

    It contains functions to override in order to catch initialisation
    of this Annotation and annotated elements binding
    (also called commonly target in the context of Annotation).

    All annotations which inherit from this are registered to target objects
    and are accessibles through the static method Annotation.get_annotations.

    Instance methods to override are:
    - __init__: set parameters during its instantiation.
    - _bind_target: called to bind target to this.
    - on_bind_target: fired when the annotated element is bound to this.

    And properties are:
    - propagate: (default True) determines if an annotation is propagated to
    all sub target elements.
    - override: (default False) exclude previous annotation of the same type as
    self class.
    - _ttl: (default None) self time to leave.
    - _in_memory: (default False) save instance in a global dictionary.

    It is also possible to set on_bind_target, propagate and override in
        the constructor.
    """

    #: attribute name for accessing to target annotations from a target
    __ANNOTATIONS_KEY__ = '__ANNOTATIONS__'

    #: attribute name for target binding notifying
    _ON_BIND_TARGET = '_on_bind_target'

    #: atribute name for target
    TARGETS = 'targets'

    #: attribute name for annotation propagatation on sub targets.
    PROPAGATE = 'propagate'

    #: attribute name for overriding base annotations.
    OVERRIDE = 'override'

    #: attribute name for annotation ttl
    TTL = '_ttl'

    #: attribute name for in_memory
    IN_MEMORY = '_in_memory'

    #: attribute name for self ts
    __TS = '_ts'

    #: attribute name for self timer
    __TIMER = '_timer'

    __slots__ = (
        _ON_BIND_TARGET, __TS, __TIMER,  # private attributes
        TARGETS, PROPAGATE, OVERRIDE, TTL, IN_MEMORY  # public attributes
    )

    # global dictionary of annotations in memory
    __ANNOTATIONS_IN_MEMORY__ = {}

    def __init__(
            self,
            on_bind_target=None, propagate=True, override=False, ttl=None,
            in_memory=False
    ):
        """Default constructor with an 'on_bind_target' handler and propagate
        scope property.

        :param on_bind_target: (None) function to call when the annotation is
            bound to an object. Function parameters are self and target.
        :param bool propagate: (True) propagate self to sub targets.
        :param bool override: (False) override old defined annotations of the
            same type.
        :param float ttl: (None) self ttl in seconds.
        :param bool in_memory: (False) save self in a global memory.
        """

        super(Annotation, self).__init__()

        # set on_bind_target handler
        setattr(self, Annotation._ON_BIND_TARGET, on_bind_target)

        # set attributes
        setattr(self, Annotation.PROPAGATE, propagate)
        setattr(self, Annotation.OVERRIDE, override)
        self.ttl = ttl
        self.in_memory = in_memory

        self.targets = set()

    def __call__(self, target, ctx=None):
        """Shouldn't be overriden by sub classes.

        :param target: target to annotate.
        :param ctx: target ctx if target is a method/function class/instance.
        :return: annotated element.
        """

        # bind target to self
        result = self.bind_target(target=target, ctx=ctx)

        return result

    def __del__(self):
        """Remove self to self.target annotations."""

        try:
            # nonify self ttl
            setattr(self, Annotation.TTL, None)

            # for all target
            for target in tuple(self.targets):
                # remove self from target
                try:
                    self.remove_from(target)

                except TypeError:
                    # raised if target is not hashable
                    pass

            # remove self from memory
            setattr(self, Annotation.IN_MEMORY, False)
        except AttributeError:
            # raised if self is deleted during __del__ execution
            pass

    @property
    def ttl(self):
        """Get actual ttl in seconds.

        :return: actual ttl.
        :rtype: float
        """

        # result is self ts
        result = getattr(self, Annotation.__TS, None)
        # if result is not None
        if result is not None:
            # result is now - result
            now = time()
            result = result - now

        return result

    @ttl.setter
    def ttl(self, value):
        """Change self ttl with input value.

        :param float value: new ttl in seconds.
        """

        # get timer
        timer = getattr(self, Annotation.__TIMER, None)

        # if timer is running, stop the timer
        if timer is not None:
            timer.cancel()

        # initialize timestamp
        timestamp = None

        # if value is None
        if value is None:
            # nonify timer
            timer = None

        else:  # else, renew a timer
            # get timestamp
            timestamp = time() + value
            # start a new timer
            timer = Timer(value, self.__del__)
            timer.start()
            # set/update attributes

        setattr(self, Annotation.__TIMER, timer)
        setattr(self, Annotation.__TS, timestamp)

    @property
    def in_memory(self):
        """
        :return: True if self is in a global memory of annotations.
        """

        self_class = self.__class__
        # check if self class is in global memory
        memory = Annotation.__ANNOTATIONS_IN_MEMORY__.get(self_class, ())
        # check if self is in memory
        result = self in memory

        return result

    @in_memory.setter
    def in_memory(self, value):
        """Add or remove self from global memory.

        :param bool value: if True(False) ensure self is(is not) in memory.
        """

        self_class = self.__class__
        memory = Annotation.__ANNOTATIONS_IN_MEMORY__

        if value:
            annotations_memory = memory.setdefault(self_class, set())
            annotations_memory.add(self)

        else:
            if self_class in memory:
                annotations_memory = memory[self_class]
                while self in annotations_memory:
                    annotations_memory.remove(self)
                if not annotations_memory:
                    del memory[self_class]

    def bind_target(self, target, ctx=None):
        """Bind self annotation to target.

        :param target: target to annotate.
        :param ctx: target ctx.
        :return: bound target.
        """

        # process self _bind_target
        result = self._bind_target(target=target, ctx=ctx)

        # fire on bind target event
        self.on_bind_target(target=target, ctx=ctx)

        return result

    def _bind_target(self, target, ctx=None):
        """Method to override in order to specialize binding of target.

        :param target: target to bind.
        :param ctx: target ctx.
        :return: bound target.
        """

        result = target

        try:
            # get annotations from target if exists.
            local_annotations = get_local_property(
                target, Annotation.__ANNOTATIONS_KEY__, [], ctx=ctx
            )
        except TypeError:
            raise TypeError('target {0} must be hashable.'.format(target))

        # if local_annotations do not exist, put them in target
        if not local_annotations:
            put_properties(
                target,
                properties={Annotation.__ANNOTATIONS_KEY__: local_annotations},
                ctx=ctx
            )

        # insert self at first position
        local_annotations.insert(0, self)

        # add target to self targets
        self.targets.add(target)

        return result

    def on_bind_target(self, target, ctx=None):
        """Fired after target is bound to self.

        :param target: newly bound target.
        :param ctx: target ctx.
        """

        _on_bind_target = getattr(self, Annotation._ON_BIND_TARGET, None)

        if _on_bind_target is not None:
            _on_bind_target(self, target=target, ctx=ctx)

    def remove_from(self, target, ctx=None):
        """Remove self annotation from target annotations.

        :param target: target from where remove self annotation.
        :param ctx: target ctx.
        """

        annotations_key = Annotation.__ANNOTATIONS_KEY__

        try:
            # get local annotations
            local_annotations = get_local_property(
                target, annotations_key, ctx=ctx
            )
        except TypeError:
            raise TypeError('target {0} must be hashable'.format(target))

        # if local annotations exist
        if local_annotations is not None:
            # if target in self.targets
            if target in self.targets:
                # remove target from self.targets
                self.targets.remove(target)
                # and remove all self annotations from local_annotations
                while self in local_annotations:
                    local_annotations.remove(self)
                # if target is not annotated anymore, remove the empty list
                if not local_annotations:
                    del_properties(target, annotations_key)

    @classmethod
    def free_memory(cls, exclude=None):
        """Free global annotation memory."""

        annotations_in_memory = Annotation.__ANNOTATIONS_IN_MEMORY__

        exclude = () if exclude is None else exclude

        for annotation_cls in list(annotations_in_memory.keys()):

            if issubclass(annotation_cls, exclude):
                continue

            if issubclass(annotation_cls, cls):
                del annotations_in_memory[annotation_cls]

    @classmethod
    def get_memory_annotations(cls, exclude=None):
        """Get annotations in memory which inherits from cls.

        :param tuple/type exclude: annotation type(s) to exclude from search.
        :return: found annotations which inherits from cls.
        :rtype: set
        """

        result = set()

        # get global dictionary
        annotations_in_memory = Annotation.__ANNOTATIONS_IN_MEMORY__

        exclude = () if exclude is None else exclude

        # iterate on annotation classes
        for annotation_cls in annotations_in_memory:

            # if annotation class is excluded, continue
            if issubclass(annotation_cls, exclude):
                continue

            # if annotation class inherits from self, add it in the result
            if issubclass(annotation_cls, cls):
                result |= annotations_in_memory[annotation_cls]

        return result

    @classmethod
    def get_local_annotations(
            cls, target, exclude=None, ctx=None, select=lambda *p: True
    ):
        """Get a list of local target annotations in the order of their
            definition.

        :param type cls: type of annotation to get from target.
        :param target: target from where get annotations.
        :param tuple/type exclude: annotation types to exclude from selection.
        :param ctx: target ctx.
        :param select: selection function which takes in parameters a target,
            a ctx and an annotation and returns True if the annotation has to
            be selected. True by default.

        :return: target local annotations.
        :rtype: list
        """

        result = []

        # initialize exclude
        exclude = () if exclude is None else exclude

        try:
            # get local annotations
            local_annotations = get_local_property(
                target, Annotation.__ANNOTATIONS_KEY__, result, ctx=ctx
            )
            if not local_annotations:
                if ismethod(target):
                    func = get_method_function(target)
                    local_annotations = get_local_property(
                        func, Annotation.__ANNOTATIONS_KEY__,
                        result, ctx=ctx
                    )
                    if not local_annotations:
                        local_annotations = get_local_property(
                            func, Annotation.__ANNOTATIONS_KEY__,
                            result
                        )
                elif isfunction(target):
                    local_annotations = get_local_property(
                        target, Annotation.__ANNOTATIONS_KEY__,
                        result
                    )

        except TypeError:
            raise TypeError('target {0} must be hashable'.format(target))

        for local_annotation in local_annotations:
            # check if local annotation inherits from cls
            inherited = isinstance(local_annotation, cls)
            # and if not excluded
            not_excluded = not isinstance(local_annotation, exclude)
            # and if selected
            selected = select(target, ctx, local_annotation)
            # if three conditions, add local annotation to the result
            if inherited and not_excluded and selected:
                result.append(local_annotation)

        return result

    @classmethod
    def remove(cls, target, exclude=None, ctx=None, select=lambda *p: True):
        """Remove from target annotations which inherit from cls.

        :param target: target from where remove annotations which inherits from
            cls.
        :param tuple/type exclude: annotation types to exclude from selection.
        :param ctx: target ctx.
        :param select: annotation selection function which takes in parameters
            a target, a ctx and an annotation and return True if the annotation
            has to be removed.
        """

        # initialize exclude
        exclude = () if exclude is None else exclude

        try:
            # get local annotations
            local_annotations = get_local_property(
                target, Annotation.__ANNOTATIONS_KEY__
            )

        except TypeError:
            raise TypeError('target {0} must be hashable'.format(target))

        # if there are local annotations
        if local_annotations is not None:

            # get annotations to remove which inherits from cls
            annotations_to_remove = [
                annotation for annotation in local_annotations
                if (
                    isinstance(annotation, cls)
                    and not isinstance(annotation, exclude)
                    and select(target, ctx, annotation)
                )
            ]

            # and remove annotations from target
            for annotation_to_remove in annotations_to_remove:
                annotation_to_remove.remove_from(target)

    @classmethod
    def get_annotations(
            cls, target, exclude=None, ctx=None, select=lambda *p: True
    ):
        """Returns all input target annotations of cls type sorted
        by definition order.

        :param type cls: type of annotation to get from target.
        :param target: target from where get annotations.
        :param tuple/type exclude: annotation types to remove from selection.
        :param ctx: target ctx.
        :param select: bool function which select annotations after applying
            previous type filters. Takes a target, a ctx and an annotation in
            parameters. True by default.
        """

        result = []

        annotations_by_ctx = get_property(
            elt=target, key=Annotation.__ANNOTATIONS_KEY__, ctx=ctx
        )

        if not annotations_by_ctx:
            if ismethod(target):
                func = get_method_function(target)
                annotations_by_ctx = get_property(
                    elt=func,
                    key=Annotation.__ANNOTATIONS_KEY__,
                    ctx=ctx
                )
                if not annotations_by_ctx:
                    annotations_by_ctx = get_property(
                        elt=func, key=Annotation.__ANNOTATIONS_KEY__
                    )
            elif isfunction(target):
                annotations_by_ctx = get_property(
                    elt=target,
                    key=Annotation.__ANNOTATIONS_KEY__
                )

        exclude = () if exclude is None else exclude

        for elt, annotations in annotations_by_ctx:

            for annotation in annotations:

                # check if annotation is a StopPropagation rule
                if isinstance(annotation, StopPropagation):
                    exclude += annotation.annotation_types

                # ensure propagation
                if elt is not target and not annotation.propagate:
                    continue

                # ensure overriding
                if annotation.override:
                    exclude += (annotation.__class__, )

                # check for annotation
                if (isinstance(annotation, cls)
                        and not isinstance(annotation, exclude)
                        and select(target, ctx, annotation)):

                    result.append(annotation)

        return result

    @classmethod
    def get_annotated_fields(cls, instance, select=lambda *p: True):
        """Get dict of {annotated fields: annotations} by cls of
        input instance.

        :return: a set of (annotated fields, annotations).
        :rtype: dict
        """

        result = {}

        for _, member in getmembers(instance):
            try:
                annotations = cls.get_annotations(
                    target=member, ctx=instance, select=select
                )

            except TypeError:
                pass

            else:
                try:
                    if annotations:
                        result[member] = annotations

                except TypeError:  # if field is an object proxy
                    pass

        return result


class StopPropagation(Annotation):
    """Stop propagation for annotation types."""

    ANNOTATION_TYPES = 'annotation_types'

    __slots__ = (ANNOTATION_TYPES,) + Annotation.__slots__

    def __init__(self, *annotation_types, **kwargs):
        """Define annotation types to not propagate at this level.
        """

        super(StopPropagation, self).__init__(**kwargs)

        setattr(self, StopPropagation.ANNOTATION_TYPES, annotation_types)


class RoutineAnnotation(Annotation):
    """Dedicated to add information on any routine, routine parameters or
    routine result."""

    #: routine attribute name
    ROUTINE = 'routine'

    #: result attribute name
    RESULT = 'result'

    #: parameters attribute name
    PARAMS = 'params'

    __slots__ = (ROUTINE, RESULT, PARAMS) + Annotation.__slots__

    def __init__(
            self, routine=None, params=None, result=None, *args, **kwargs
    ):
        """
        :param routine: routine information.
        :param params: params information.
        :param result: routine result information.
        """

        super(RoutineAnnotation, self).__init__(*args, **kwargs)

        setattr(self, RoutineAnnotation.ROUTINE, routine)
        setattr(self, RoutineAnnotation.PARAMS, params)
        setattr(self, RoutineAnnotation.RESULT, result)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Decorators dedicated to class or functions calls."""

from __future__ import absolute_import

from .interception import PrivateInterceptor
from .check import Target

from b3j0f.utils.iterable import first
from b3j0f.utils.version import getcallargs

from six import get_function_code
from six.moves import range

from sys import stderr, maxsize

from time import sleep

from functools import wraps

__all__ = ['Types', 'types', 'Curried', 'curried', 'Retries', 'Memoize']


@Target(callable)
class Types(PrivateInterceptor):
    """
    Check routine parameters and return type.
    """

    class TypesError(Exception):
        """Handle Types error."""

    class SpecialCondition(object):
        """Handle SpecialCondition."""

        def __init__(self, _type):

            super(Types.SpecialCondition, self).__init__()

            self._type = _type

        def get_type(self):
            """Get special condition parameter type."""

            return self._type

    class NotNone(SpecialCondition):
        """Handle NotNone SpecialCondition."""

    class NotEmpty(SpecialCondition):
        """Handle NotEmpty SpecialCondition."""

    class NamedParameterType(object):
        """Handle Named Parameter Type."""

        def __init__(self, name, parameter_type):

            super(Types.NamedParameterType, self).__init__()

            self._name = name
            self._parameter_type = parameter_type

    class NamedParameterTypes(object):
        """Handle Named Parameter Types."""

        def __init__(self, target, named_parameter_types):

            super(Types.NamedParameterTypes, self).__init__()

            self._named_parameter_types = []

            target_code = get_function_code(target)

            for index in range(target_code.co_argcount):
                targetparamname = target_code.co_varnames[index]

                if targetparamname in named_parameter_types:
                    parameter_type = named_parameter_types[targetparamname]
                    named_parameter_type = Types.NamedParameterType(
                        targetparamname,
                        parameter_type
                    )
                    self._named_parameter_types.append(named_parameter_type)

                else:
                    self._named_parameter_types.append(None)

    #: return type attribute name
    RTYPE = 'rtype'

    #: parameter types attribute name
    PTYPES = 'ptypes'

    __slots__ = (RTYPE, PTYPES) + PrivateInterceptor.__slots__

    """
    Check parameter or result types of decorated class or function call.
    """
    def __init__(self, rtype=None, ptypes=None, *args, **kwargs):
        """
        :param rtype:
        """

        super(Types, self).__init__(*args, **kwargs)

        self.rtype = rtype
        self.ptypes = {} if ptypes is None else ptypes

    @staticmethod
    def check_value(value, expected_type):
        """Check Types parameters."""

        result = False

        if isinstance(expected_type, Types.NotNone):
            result = value is not None and Types.check_value(
                value,
                expected_type.get_type()
            )

        else:
            result = value is None

            if not result:

                value_type = type(value)

                if isinstance(expected_type, Types.NotEmpty):
                    try:
                        result = len(value) != 0
                        if result:
                            _type = expected_type.get_type()
                            result = Types.check_value(value, _type)
                    except TypeError:
                        result = False

                elif isinstance(expected_type, list):
                    result = issubclass(value_type, list)

                    if result:
                        if len(expected_type) == 0:
                            result = len(value) == 0
                        else:
                            _expected_type = expected_type[0]

                            for item in value:
                                result = Types.check_value(
                                    item,
                                    _expected_type
                                )

                                if not result:
                                    break

                elif isinstance(expected_type, set):
                    result = issubclass(value_type, set)

                    if result:
                        if len(expected_type) == 0:
                            result = len(value) == 0
                        else:

                            _expected_type = expected_type.copy().pop()
                            _value = value.copy()

                            value_length = len(_value)

                            for _ in range(value_length):
                                item = _value.pop()
                                result = Types.check_value(
                                    item,
                                    _expected_type)

                                if not result:
                                    break
                else:
                    result = issubclass(value_type, expected_type)

        return result

    def _interception(self, joinpoint):

        target = joinpoint.target
        args = joinpoint.args
        kwargs = joinpoint.kwargs

        if self.ptypes:
            callargs = getcallargs(target, *args, **kwargs)

            for arg in callargs:
                value = callargs[arg]
                expected_type = self.ptypes.get(arg)

                if (
                        expected_type is not None and
                        not Types.check_value(value, expected_type)
                ):
                    raise Types.TypesError(
                        "wrong typed parameter for arg {0} : {1} ({2}). \
                        Expected: {3}."
                        .format(
                            arg, value, type(value), expected_type
                        )
                    )

        result = joinpoint.proceed()

        target = joinpoint.target
        args = joinpoint.args
        kwargs = joinpoint.kwargs

        if self.rtype:
            if not Types.check_value(result, self.rtype):
                raise Types.TypesError(
                    "wrong result type for {0} with parameters {1}, {2}: {3} \
                    ({4}). Expected {5}."
                    .format(
                        target, args, kwargs, result, type(result),
                        self.rtype
                    )
                )

        return result


def types(*args, **kwargs):
    """Quick alias for the Types Annotation with only args and kwargs
    parameters.

    :param tuple args: may contain rtype.
    :param dict kwargs: may contain ptypes.
    """

    rtype = first(args)

    return Types(rtype=rtype, ptypes=kwargs)


class Curried(PrivateInterceptor):
    """Annotation that returns a function that keeps returning functions
    until all arguments are supplied; then the original function is
    evaluated.

    Inspirated from Jeff Laughlin Consulting LLC projects.
    """

    ARGS = 'args'  #: args attribute name
    KWARGS = 'kwargs'  #: kwargs attribute name
    DEFAULT_ARGS = 'default_args'  #: default args attribute name
    DEFAULT_KWARGS = 'default_kwargs'  #: default kwargs attribute name

    __slots__ = (
        ARGS, KWARGS, DEFAULT_ARGS, DEFAULT_KWARGS
    ) + PrivateInterceptor.__slots__

    class CurriedResult(object):
        """Curried result in case of missing arguments."""

        __slots__ = ('curried', 'exception')

        def __init__(self, curried, exception):

            super(Curried.CurriedResult, self).__init__()

            self.curried = curried
            self.exception = exception

    def __init__(self, varargs=None, keywords=None, *args, **kwargs):
        """
        :param tuple varargs: function call varargs.
        :param dict keywords: function call keywords.
        """

        super(Curried, self).__init__(*args, **kwargs)

        # initialize arguments
        if varargs is None:
            varargs = ()
        if keywords is None:
            keywords = {}
        # set attributes
        self.args = self.default_args = varargs
        self.kwargs = self.default_kwargs = keywords

    def _bind_target(self, target, *args, **kwargs):

        @wraps(target)
        def wrapper(*args, **kwargs):
            """Target wrapper."""

            return target(*args, **kwargs)

        result = super(Curried, self)._bind_target(
            target=wrapper, *args, **kwargs
        )

        return result

    def _interception(self, joinpoint, *args, **kwargs):

        result = None

        target = joinpoint.target

        args = joinpoint.args
        kwargs = joinpoint.kwargs

        self.kwargs.update(kwargs)
        self.args += args

        try:
            # check if all arguments are given
            getcallargs(target, *self.args, **self.kwargs)
            joinpoint.args = self.args
            joinpoint.kwargs = self.kwargs
            result = joinpoint.proceed()
        except TypeError as ex:
            # in case of problem, returns curried decorater and exception
            result = Curried.CurriedResult(self, ex)

        return result


def curried(*args, **kwargs):
    """Curried annotation with varargs and kwargs.
    """

    return Curried(varargs=args, keywords=kwargs)


def example_exc_handler(tries_remaining, exception, delay):
    """Example exception handler; prints a warning to stderr.

    tries_remaining: The number of tries remaining.
    exception: The exception instance which was raised.
    """

    print >> stderr, "Caught '{0}', {1} tries remaining, \
    sleeping for {2} seconds".format(exception, tries_remaining, delay)


class Retries(PrivateInterceptor):
    """Function decorator implementing retrying logic.

    condition: retry condition, among execution success or failure or both.
    delay: Sleep this many seconds * backoff * try number after failure
    backoff: Multiply delay by this factor after each failure
    exceptions: A tuple of exception classes; default (Exception,)
    hook: A function with the signature myhook(data, condition, tries_remaining
    , mydelay) where data is result function or raised Exception, condition is
    ON_ERROR or ON_SUCCESS depending on error or success execution function,
    tries_remaining is tries remaining, and finally, mydelay is waiting seconds
    between calls; default None.

    The decorator will call the function up to max_tries times if it raises
    an exception or if it simply execute the function, depending on state
    condition.

    By default it catches instances of the Exception class and subclasses.
    This will recover after all but the most fatal errors. You may specify a
    custom tuple of exception classes with the 'exceptions' argument; the
    function will only be retried if it raises one of the specified
    exceptions.

    Additionally you may specify a hook function which will be called prior
    to retrying with the number of remaining tries and the exception instance;
    see given example. This is primarily intended to give the opportunity to
    log the failure. Hook is not called after failure if no retries remain.
    """

    MAX_TRIES = 'max_tries'  #: max_tries attribute name.
    DELAY = 'delay'  #: delay attribute name.
    BACKOFF = 'backoff'  #: backoff attribute name.
    EXCEPTIONS = 'exceptions'  #: exceptions attribute name.
    HOOK = 'hook'  #: hook attribute name.
    CONDITION = 'condition'  #: condition attribute name.

    DEFAULT_DELAY = 1
    DEFAULT_BACKOFF = 2
    DEFAULT_EXCEPTIONS = (Exception,)

    ON_ERROR = 1  #: on error retries condition.
    ON_SUCCESS = 2  #: on success retries condition.
    ALL = ON_ERROR | ON_SUCCESS  #: all retries condition.

    __slots__ = (
        MAX_TRIES, DELAY, BACKOFF, EXCEPTIONS, HOOK, CONDITION
    ) + PrivateInterceptor.__slots__

    def __init__(
            self,
            max_tries,
            delay=DEFAULT_DELAY,
            backoff=DEFAULT_BACKOFF,
            exceptions=DEFAULT_EXCEPTIONS,
            hook=None,
            condition=ALL,
            *args, **kwargs
    ):

        super(Retries, self).__init__(*args, **kwargs)

        self.max_tries = max_tries
        self.delay = delay
        self.backoff = backoff
        self.exceptions = exceptions
        self.hook = hook
        self.condition = condition

    def _interception(self, joinpoint):

        result = None

        mydelay = self.delay

        for tries_remaining in range(self.max_tries - 1, -1, -1):

            try:
                result = joinpoint.proceed()

            except self.exceptions as ex:

                mydelay = self._checkretry(
                    mydelay=mydelay, condition=Retries.ON_ERROR, data=ex,
                    tries_remaining=tries_remaining
                )

            else:

                mydelay = self._checkretry(
                    mydelay=mydelay, condition=Retries.ON_SUCCESS, data=result,
                    tries_remaining=tries_remaining
                )

                if mydelay is None:
                    break  # stop execution if mydelay is None

        return result

    def _checkretry(self, mydelay, condition, tries_remaining, data):
        """Check if input parameters allow to retries function execution.

        :param float mydelay: waiting delay between two execution.
        :param int condition: condition to check with this condition.
        :param int tries_remaining: tries remaining.
        :param data: data to hook.
        """

        result = mydelay

        if self.condition & condition and tries_remaining > 0:

            # hook data with tries_remaining and mydelay
            if self.hook is not None:
                self.hook(data, condition, tries_remaining, mydelay)
            # wait mydelay seconds
            sleep(mydelay)
            result *= self.backoff  # increment mydelay with this backoff

        elif condition is Retries.ON_ERROR:
            raise data  # raise data if no retries and on_error

        else:  # else Nonify mydelay to prevent callee function to stop
            result = None

        return result


class Memoize(PrivateInterceptor):
    """Save funtion results related to called parameters.

    Parameters must be hashable."""

    MAX_SIZE = 'max_size'  #: max size result.
    _CACHE = '_cache'  #: cache object which stores results and params.

    DEFAULT_MAX_SIZE = maxsize  #: default max size value.

    __slots__ = (MAX_SIZE, _CACHE) + PrivateInterceptor.__slots__

    def __init__(self, max_size=DEFAULT_MAX_SIZE, *args, **kwargs):

        super(Memoize, self).__init__(*args, **kwargs)

        self.max_size = max_size
        self._cache = {}

    def _getkey(self, args, kwargs):
        """Get hash key from args and kwargs.

        args and kwargs must be hashable.

        :param tuple args: called vargs.
        :param dict kwargs: called keywords.
        :return: hash(tuple(args) + tuple((key, val) for key in sorted(kwargs)).
        :rtype: int."""

        values = list(args)

        keys = sorted(list(kwargs))

        for key in keys:
            values.append((key, kwargs[key]))

        result = hash(tuple(values))

        return result

    def _interception(self, joinpoint):

        result = None

        args = joinpoint.args
        kwargs = joinpoint.kwargs
        key = self._getkey(args, kwargs)

        _cache = self._cache

        if key in _cache:
            _, _, result = _cache[key]

        else:
            result = joinpoint.proceed()

            if len(self._cache) < self.max_size:
                _cache[key] = (args, kwargs, result)

        return result

    def getparams(self, result):
        """Get result parameters.

        :param result: cached result.
        :raises: ValueError if result is not cached.
        :return: args and kwargs registered with input result.
        :rtype: tuple"""

        for key in self._cache:
            if self._cache[key][2] == result:
                args, kwargs, _ = self._cache[key]
                return args, kwargs
                break

        else:
            raise ValueError('Result is not cached')

    def clearcache(self):
        """Clear cache."""

        self._cache.clear()
